import yaml

from databutton.utils import get_project_directory

from .utils import base64str_to_str, str_to_base64str


class SecretsFileClient:
    def __init__(self, secrets_filename=None):
        if secrets_filename is None:
            secrets_filename = get_project_directory() / "databutton-secrets.yaml"
        self._secrets_filename = secrets_filename

    def _read_local_secrets(self) -> dict:
        try:
            # Could also support a secrets directory with
            # files named as keys and containing the values,
            # that's what we'll do for the docker mounting thing?
            with open(self._secrets_filename, "r") as f:
                secrets = yaml.safe_load(f)
            return secrets or {}
        except FileNotFoundError:
            return {}

    def add(self, name: str, value: str) -> bool:
        secrets = self._read_local_secrets()
        secrets[name] = str_to_base64str(value)
        with open(self._secrets_filename, "w") as f:
            yaml.safe_dump(secrets, f)
        return True

    def delete(self, name: str) -> bool:
        secrets = self._read_local_secrets()
        if name in secrets:
            del secrets[name]
        with open(self._secrets_filename, "w") as f:
            yaml.safe_dump(secrets, f)
        return True

    def get(self, name: str) -> str:
        value_base64 = self._read_local_secrets()[name]
        return base64str_to_str(value_base64)

    def list(self) -> list[str]:
        return sorted(self._read_local_secrets().keys())



import contextlib
import threading
import time

import uvicorn


class Server(uvicorn.Server):
    is_finished = False

    def install_signal_handlers(self):
        pass

    @contextlib.contextmanager
    def run_in_thread(self):
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()


def get_threaded_uvicorn(config: uvicorn.Config):
    return Server(config=config)



#
# Module containing helpers to scan through files
#
import fnmatch
import os
import re
from pathlib import Path

from databutton.utils import DEFAULT_GLOB_EXCLUDE, get_databutton_config

# Find all modules we need to import
# 1. Traverse all sub-directories under rootdir
# 2. Find all python files containing a decorator
# 3. If __init__.py doesn't exist in a subdir with a decorator, create it
# 4. Rename python filename to an importable module name
# 5. Return list of modules to import


def find_databutton_directive_modules(rootdir=Path.cwd()):
    try:
        config = get_databutton_config()
    except FileNotFoundError:
        config = None
    excludes = config.exclude if config else DEFAULT_GLOB_EXCLUDE
    modules_to_import = []
    for root, dirs, files in os.walk(rootdir):
        dirs[:] = [d for d in dirs if d not in excludes]
        for file in files:
            filepath_norm = Path(root, file)
            file_norm = Path(file)
            should_exclude = False
            for exclude in excludes:
                if fnmatch.fnmatch(file_norm, exclude):
                    # Skip this file
                    should_exclude = True
                    break
            if should_exclude:
                continue
            if file.endswith(".py"):
                try:
                    s = open(filepath_norm, "r").read()
                except UnicodeDecodeError as e:
                    print(f"Could not read file {file}")
                    print(e)
                    continue
                decorators = ["apps.streamlit", "jobs.repeat_every"]
                for decorator in decorators:
                    occurences = len(re.findall(rf"@[\w]+.{decorator}", s))
                    if occurences > 0:
                        if root != rootdir:
                            init_file = Path(root) / "__init__.py"
                            if not init_file.exists():
                                open(init_file, "a").close()
                        md = filepath_norm
                        md = Path(md.as_posix().replace("./", ""))
                        md = md.as_posix().replace(".py", "").replace("/", ".")
                        modules_to_import.append(str(md))

    return modules_to_import



def get_secrets_client():
    from .apiclient import SecretsApiClient

    return SecretsApiClient()


def get(name: str) -> str:
    """Get project secret."""
    return get_secrets_client().get(name)



import datetime
import logging

import requests

from databutton.utils import get_auth_token, get_cached_databutton_config

FIRESTORE_BASE_URL = "https://firestore.googleapis.com/v1/projects/databutton/databases/(default)/documents"

logger = logging.getLogger("databutton.scheduler")


def create_job_run(job_id: str, run_id: str, start_time: str):
    try:
        project_id = get_cached_databutton_config().uid
        res = requests.post(
            f"{FIRESTORE_BASE_URL}/projects/{project_id}/runs",
            params={"documentId": run_id},
            headers={"Authorization": f"Bearer {get_auth_token()}"},
            json={
                "fields": {
                    "jobId": {"stringValue": job_id},
                    "startTime": {"timestampValue": start_time},
                }
            },
        )
        if res.ok:
            data = res.json()
            return data
        else:
            logger.error("Could not create job run")
            logger.error(res.text)
        return None
    except Exception:
        import traceback

        logger.error(traceback.format_exc())
        return None


def update_job_run(
    job_id: str,
    run_id: str,
    start_time: str,
    end_time: str,
    next_run_time: str,
    success: bool,
):
    try:
        project_id = get_cached_databutton_config().uid
        start = datetime_from_iso_utc_timestamp(start_time)
        end = datetime_from_iso_utc_timestamp(end_time)
        duration = int((end - start).microseconds / 1000)
        fields = {
            "jobId": {"stringValue": job_id},
            "startTime": {"timestampValue": start_time},
            "endTime": {"timestampValue": end_time},
            "nextRunTime": {"timestampValue": next_run_time},
            "duration": {"integerValue": duration},
            "success": {"booleanValue": success},
        }
        res = requests.patch(
            f"{FIRESTORE_BASE_URL}/projects/{project_id}/runs/{run_id}",
            headers={"Authorization": f"Bearer {get_auth_token()}"},
            json={
                "fields": fields,
            },
        )
        if res.ok:
            data = res.json()
            return data
        else:
            logger.error("Could not update job run")
            logger.error(res.text)
        return None
    except Exception:
        import traceback

        logger.error(traceback.format_exc())
        return None


def utc_now() -> datetime.datetime:
    return datetime.datetime.utcnow()


def iso_utc_timestamp_now() -> str:
    return utc_now().isoformat() + "Z"


def iso_utc_timestamp(dt: datetime.datetime) -> str:
    return (
        datetime.datetime.fromtimestamp(dt.timestamp(), tz=datetime.timezone.utc)
        .replace(tzinfo=None)
        .isoformat()
        + "Z"
    )


def datetime_from_iso_utc_timestamp(ts: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(ts.removesuffix("Z")).astimezone(
        datetime.timezone.utc
    )


def push_job_log(run_id: str, msg: bool):
    try:
        project_id = get_cached_databutton_config().uid
        res = requests.post(
            f"{FIRESTORE_BASE_URL}/projects/{project_id}/runs/{run_id}/logs",
            headers={"Authorization": f"Bearer {get_auth_token()}"},
            json={
                "fields": {
                    "time": {"timestampValue": iso_utc_timestamp_now()},
                    "msg": {"stringValue": msg},
                }
            },
        )
        if res.ok:
            data = res.json()
            return data
        else:
            logger.error(res.text, res.status_code)
        return None
    except Exception:
        import traceback

        logger.error(traceback.format_exc())
        return None



import logging

import trio


class Tracer(trio.abc.Instrument):
    def __init__(self, logger_name=None):
        super().__init__()
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        else:
            self.logger = logging.getLogger()

    def _print_with_task(self, msg, task):
        # repr(task) is perhaps more useful than task.name in general,
        # but in context of a tutorial the extra noise is unhelpful.
        self.logger.debug(f"{msg}: {task.name}")

    def task_spawned(self, task):
        self._print_with_task("### new task spawned", task)

    def task_scheduled(self, task):
        self._print_with_task("### task scheduled", task)

    def before_task_step(self, task):
        self._print_with_task(">>> about to run one step of task", task)

    def after_task_step(self, task):
        self._print_with_task("<<< task step finished", task)

    def task_exited(self, task):
        self._print_with_task("### task exited", task)



import os
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from databutton.utils import ProjectConfig, get_databutton_config
from databutton.version import __version__

if TYPE_CHECKING:
    from databutton.utils.build import ArtifactDict


def generate_kube_resources(cpu: str, memory: str):
    return {
        "requests": {"memory": memory, "cpu": cpu},
        "limits": {"memory": memory, "cpu": cpu},
    }


kube_request_resources = generate_kube_resources(cpu="1", memory="2056Mi")

KUBE_HEALTHCHECK = {
    "livenessProbe": {
        "httpGet": {"path": "/", "port": "http"},
    },
    "readinessProbe": {"httpGet": {"path": "/", "port": "http"}},
}

_streamlit_args = [
    "--server.address=0.0.0.0",
    "--server.fileWatcherType=none",
    "--server.enableWebsocketCompression=true",
    "--browser.gatherUsageStats=false",
    "--global.dataFrameSerialization=arrow",
    "--server.headless=true",
    "--server.enableCORS=false",
    "--server.enableXsrfProtection=false",
]


def generate_manifest(
    artifacts: "ArtifactDict", yaml_path=Path(".databutton", "kubernetes")
):
    try:
        config = get_databutton_config()
    except Exception:
        config = ProjectConfig(uid="not-set", name="not-set")
    project_id = config.uid

    env_from_secret = {"envFrom": [{"secretRef": {"name": "databutton-secret"}}]}

    IMAGE_NAME = os.environ.get("IMAGE_NAME", "databutton-base")

    streamlit_apps = []
    for app in artifacts.streamlit_apps:
        st_app = {
            "name": f"route-{app.uid}",
            "image": IMAGE_NAME,
            "imagePullPolicy": "IfNotPresent",
            "args": [
                app.filename,
                "--server.port",
                f"{app.port}",
            ]
            + _streamlit_args,
            "ports": [{"containerPort": app.port, "name": "http"}],
            "env": [{"name": "PYTHONPATH", "value": "."}],
            **env_from_secret,
            "command": ["streamlit", "run"],
            **KUBE_HEALTHCHECK,
        }
        st_app["resources"] = generate_kube_resources(cpu=app.cpu, memory=app.memory)
        streamlit_apps.append(st_app)

    namespace = {
        "apiVersion": "v1",
        "kind": "Namespace",
        "metadata": {"name": project_id},
    }

    kustomization = {
        "apiVersion": "kustomize.config.k8s.io/v1beta1",
        "kind": "Kustomization",
        "namespace": project_id,
        "resources": [
            "namespace.yaml",
            "deployment.yaml",
            "service.yaml",
            "mapping.yaml",
        ],
        "commonLabels": {
            "databutton.com/project_id": project_id,
        },
        "commonAnnotations": {
            "databutton.com/version": __version__,
        },
    }

    scheduler_container = {
        "name": "scheduler",
        "image": IMAGE_NAME,
        "imagePullPolicy": "IfNotPresent",
        "command": ["databutton", "schedule"],
        "env": [{"name": "PYTHONPATH", "value": "."}],
        **env_from_secret,
        "resources": generate_kube_resources(cpu="1", memory="2056Mi"),
    }

    deployment = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {
            "name": "databutton",
        },
        "spec": {
            "replicas": 1,
            "template": {
                "spec": {
                    "containers": streamlit_apps
                    + [scheduler_container]
                    + [
                        {
                            "name": "main",
                            "image": IMAGE_NAME,
                            "ports": [{"containerPort": 8000, "name": "http"}],
                            "imagePullPolicy": "IfNotPresent",
                            "command": ["databutton", "serve"],
                            "env": [
                                {"name": "PYTHONPATH", "value": "."},
                            ],
                            **env_from_secret,
                            "resources": generate_kube_resources(
                                cpu="0.2", memory="256Mi"
                            ),
                            **KUBE_HEALTHCHECK,
                        }
                    ]
                },
            },
        },
    }

    mapping = {
        "apiVersion": "getambassador.io/v3alpha1",
        "kind": "Mapping",
        "metadata": {"name": "databutton"},
        "spec": {
            "hostname": f"p{project_id}.dbtn.app",
            "service": "databutton",
            "prefix": "/",
            "allow_upgrade": ["websocket"],
        },
    }

    service = {
        "apiVersion": "v1",
        "kind": "Service",
        "metadata": {"name": "databutton"},
        "spec": {"ports": [{"port": 80, "targetPort": 8000, "name": "http"}]},
    }

    files = {
        "deployment": deployment,
        "service": service,
        "mapping": mapping,
        "kustomization": kustomization,
        "namespace": namespace,
    }

    if yaml_path is not None:
        kustomize_path = yaml_path
        kustomize_path.mkdir(exist_ok=True, parents=True)
        for kind, d in files.items():
            with open(kustomize_path / f"{kind}.yaml", "w") as f:
                yaml.safe_dump(d, f, allow_unicode=True)

    return files



import logging
import os
import shutil
import sys
import tarfile
import tempfile
import webbrowser
from pathlib import Path
from time import sleep
from uuid import uuid4

import click
import requests
import trio
from alive_progress import alive_bar
from uvicorn import Config

from databutton.decorators.jobs.schedule import Scheduler
from databutton.secrets.apiclient import SecretsApiClient
from databutton.secrets.client import get_secrets_client
from databutton.server.dev import DatabuttonRunner
from databutton.utils import (
    LoginData,
    create_databutton_cloud_project,
    create_databutton_config,
    get_api_url,
    get_auth_token,
    get_build_logs,
    get_databutton_config,
    get_databutton_login_info,
    get_databutton_login_path,
    new_databutton_version_exists,
)
from databutton.utils.deploy import (
    create_archive,
    get_build_id_from_deployment,
    listen_to_build,
    upload_archive,
)
from databutton.utils.uvicorn_in_mem import get_threaded_uvicorn
from databutton.version import __version__

logger = logging.getLogger("databutton.cli")

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG,
}


def require_databutton_login_info() -> LoginData:
    login_info = get_databutton_login_info()
    if login_info is None:
        click.secho("❌ You're not authenticated to Databutton, yet.")
        click.secho(
            f"Type {click.style('databutton login', fg='cyan')} to login or create a user."
        )
        exit(1)
    return login_info


@click.group()
@click.option("--verbose", "-v", count=True, help="Enable verbose logging")
@click.pass_context
def cli(ctx, verbose: int):
    """Run databutton."""
    ctx.ensure_object(dict)
    # Use the verbosity count to determine the logging level...
    if verbose > 0:
        logger.setLevel(
            LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else logging.DEBUG
        )
    ctx.obj["VERBOSE"] = verbose


@cli.command()
def version():
    """Get the library version."""
    click.echo(click.style(f"{__version__}", bold=True))


@cli.command()
@click.argument(
    "project-directory",
    required=True,
)
@click.option(
    "-t",
    "--template",
    required=False,
    default="hello-databutton",
    help="A template to bootstrap off of.",
)
@click.option(
    "-n",
    "--name",
    required=False,
    help="The name of your Databutton project (on databutton.com)",
)
@click.option("--non-interactive", default=False, is_flag=True)
@click.pass_context
def create(
    ctx: click.Context,
    project_directory: Path,
    template: str,
    name: str,
    non_interactive: bool,
):
    """
    Create a Databutton project in the provided project-directory

    PROJECT_DIRECTORY is the name of the directory for your Databutton project.

    Examples:

    \b
    # Create a new Databutton project in the `my-new-project` directory
    databutton create my-new-project

    \b
    # Create a new Databutton project with a custom name
    databutton create my-new-project -n "My new Databutton project"

    \b
    # Create a new Databutton project in the current directory
    databutton create .

    \b
    # Create a new Databutton project from a template
    databutton create my-new-project --template=sales-forecasting

    """
    project_directory = Path(project_directory).resolve()
    project_name = name if name else str(project_directory.as_posix()).split("/")[-1]

    if project_directory.exists():
        if len(list(project_directory.iterdir())) > 0:
            click.secho("❌ The target directory isn't empty.")
            exit(1)
    login_info = get_databutton_login_info()
    if not login_info:
        click.echo()
        click.secho(
            "It doesn't seem like you have a user with Databutton, at least you're not logged in."
        )
        if non_interactive:
            click.echo(
                "Can't log interactively in non-interactive mode."
                + "\nRun databutton login first or run databutton create without --non-interactive"
            )
            exit(1)
        click.secho("Let's log you in or create a user...")
        click.echo()
        ctx.invoke(login)
        login_info = get_databutton_login_info()
    if not login_info:
        # This might still not be true...
        click.echo("Could not log in successfully, please try again..")
        exit(1)
    project_id = create_databutton_cloud_project(project_name)

    click.secho("Downloading template...")
    res = requests.get(
        f"https://storage.googleapis.com/databutton-app-templates/{template}.tar.gz",
        stream=True,
    )
    if not res.ok:
        click.secho(
            f"❌ Failed to get template {template}, are you sure you typed an existing template?"
        )
        exit(1)
    with tempfile.TemporaryFile() as tmpfile:
        tmpfile.write(res.raw.read())
        tmpfile.seek(0)
        with tarfile.open(fileobj=tmpfile) as t:
            t.extractall(project_directory)
    create_databutton_config(
        name=project_name, uid=project_id, project_directory=project_directory
    )
    click.echo()
    click.echo(
        f"Success! Created project {click.style(project_name, fg='cyan')} in {project_directory}"
    )
    click.echo("Inside that directory, you can run several commands:")
    click.echo()

    commands = [
        ("databutton start", "Starts the development server."),
        (
            "databutton deploy",
            "Deploy your project to databutton.com.\n    "
            + "Note that in order to deploy you need a user and a project in Databutton.",
        ),
        (
            "databutton build",
            "Bundles the project and generates the necessary files for production.",
        ),
    ]

    for cmd, desc in commands:
        click.echo(f"  {click.style(cmd, fg='cyan')}")
        click.echo(f"    {click.style(desc)}")
        click.echo()

    click.echo("We suggest you begin by typing:\n")

    cd_dir = (
        project_directory.relative_to(Path.cwd())
        if project_directory.is_relative_to(Path.cwd())
        else project_directory
    )

    click.echo(f"  {click.style('cd', fg='cyan')} {cd_dir}")
    click.secho("  pip install -r requirements.txt", fg="cyan")
    click.secho("  databutton start", fg="cyan")
    click.secho("  databutton deploy", fg="cyan")

    click.echo()
    click.secho(
        f"You can always see the available commands by typing {click.style('databutton --help', fg='cyan')}."
        + f"\nDocumentation is available by typing {click.style('databutton docs', fg='cyan')}."
    )

    click.echo("\nHappy building!\n")


@cli.command()
@click.option("--debug", default=False, type=bool, show_default=True, is_flag=True)
def start(debug=False):
    """Run the Databutton development server"""
    new_version = new_databutton_version_exists()
    if new_version:
        click.echo()
        click.echo("There is a new version of Databutton available")
        click.echo()
        if Path("pyproject.toml").exists():
            # Poetry project
            click.echo(
                f'write {click.style(f"poetry add databutton@{new_version}", fg="yellow")} to install the new version.'
            )
        else:
            click.echo(
                f'write {click.style(f"pip install databutton=={new_version}", fg="yellow")} to install the new version'
            )
        click.echo()
        click.echo()

    # Make sure a config exists
    try:
        get_databutton_config()
    except FileNotFoundError:
        click.secho("Could not find a databutton.json file.")
        click.secho(
            f"Are you sure you ran {click.style('databutton start', fg='cyan')} in the right directory?"
        )
        exit(1)
    except Exception as e:
        print(e)
        exit(1)

    runner = DatabuttonRunner()

    try:
        trio.run(
            runner.run,
            debug,
            # instruments=[Tracer(logger_name="databutton.start")],
        )
    except KeyboardInterrupt:
        logger.debug("Closing scheduler")
        sys.exit(0)


@cli.command()
def login():
    """Login to Databutton"""
    click.echo(click.style("Opening browser to authenticate.."))

    dir_path = Path(__file__).parent
    app_dir = dir_path / "auth"
    sys.path.insert(0, str(app_dir.resolve()))
    login_url = "http://localhost:8008"
    server = get_threaded_uvicorn(
        Config("server:app", port=8008, log_level="error", reload=False)
    )

    # Remove login details
    shutil.rmtree(get_databutton_login_path(), ignore_errors=True)
    with server.run_in_thread():
        webbrowser.open(login_url)
        login_path = get_databutton_login_path()
        while not (os.path.exists(login_path) and len(os.listdir(login_path)) > 0):
            sleep(1)
        # Wait a little bit so the user gets a response
    click.secho("Logged in!")


@cli.command()
@click.pass_context
def deploy(ctx: click.Context):
    """Deploy your project to Databutton"""

    _ = require_databutton_login_info()

    try:
        config = get_databutton_config()
    except FileNotFoundError:
        click.secho(
            "❌ Can't find a Databutton config file. Are you in the correct folder?"
        )
        click.secho(
            f"Type {click.style('databutton init', fg='cyan')} to "
            + "create a new Databutton project in this directory."
        )
        exit(1)

    deployment_id = str(uuid4())

    click.echo()
    click.echo(click.style(f"=== Deploying to {config.name}", fg="green"))
    # ctx.invoke(build)
    click.echo(click.style("i packaging components...", fg="cyan"))

    with tempfile.NamedTemporaryFile() as tmpfile:
        create_archive(tmpfile, source_dir=Path.cwd(), config=config)
        click.echo(click.style("i done packaging components", fg="green"))
        click.echo(click.style("i uploading components", fg="cyan"))
        try:
            upload_archive(config, deployment_id, tmpfile)
        except Exception:
            click.secho("❌ Could not upload, talk to someone at Databutton to debug...")
            exit(1)

        click.echo(click.style("i finished uploading components", fg="green"))
        click.echo(click.style("i cleaning up", fg="cyan"))

    click.echo(
        click.style(
            "i waiting for deployment to be ready, this can take a few minutes...",
            fg="cyan",
        )
    )
    click.secho("i deploying...", fg="cyan")
    click.echo()
    click.secho(
        "i you can close this window if you want, the deploy will continue in the cloud.",
        fg="cyan",
    )

    build_id = get_build_id_from_deployment(deployment_id)
    build_logs = get_build_logs(build_id)
    click.echo()
    click.echo(f"Build logs are available at {build_logs}")
    click.echo()

    with alive_bar(
        stats=False, monitor=False, monitor_end=False, elapsed=False, title=""
    ) as bar:
        for status in listen_to_build(deployment_id):
            bar.text = f"status: {status.capitalize() if status else 'Queued'}"
            bar()

    if status == "SUCCESS":
        click.echo(click.style("✅ Done!", fg="green"))
        click.echo()
        styled_url = click.style(
            f"https://next.databutton.com/projects/{config.uid}", fg="cyan"
        )
        click.secho(f"You can now go to \n\t{styled_url}")
        click.echo()
    elif status == "FAILURE":
        click.echo(click.style("❌ Error deploying...", fg="red"))
        click.echo()
    elif status == "CANCELLED":
        click.secho(
            "Your build was cancelled. A databutler probably did that on purpose."
        )
        click.secho("You should reach out!")


@cli.command()
def build():
    """Build the project, built components will be found in .databutton"""
    click.echo(click.style("i building project", fg="cyan"))
    from databutton.utils.build import generate_components

    click.echo(click.style("i generating components", fg="cyan"))
    artifacts = generate_components()
    click.echo(click.style("i finished building project in .databutton", fg="green"))
    return artifacts


@cli.command()
def serve():
    """Starts a web server for production."""
    click.echo(click.style("=== Serving"))
    import uvicorn

    port = os.environ["PORT"] if "PORT" in os.environ else 8000
    uvicorn.run(
        "databutton.server.prod:app",
        reload=False,
        port=int(port),
        host="0.0.0.0",
    )


@cli.command("init")
@click.option("--name", help="Name of the project")
def init(name: str):
    """Creates a new project in Databutton and writes to databutton.json"""

    _ = require_databutton_login_info()

    if os.path.exists("databutton.json"):
        click.secho("There is already a databutton.json file in this directory")
        project_config = get_databutton_config()
        click.secho(f"  id: {click.style(project_config.uid, fg='cyan')}")
        click.secho(f"  name: {click.style(project_config.name, fg='cyan')}")
        should_overwrite = click.confirm(
            "Do you want to create a new project and overwrite it?", default=True
        )
        if not should_overwrite:
            click.secho("You did not create a new Databutton project.")
            exit(0)
    token = get_auth_token()
    if not name:
        name = click.prompt("Choose a name for your databutton project", type=str)
    res = requests.post(
        "https://europe-west1-databutton.cloudfunctions.net/createOrUpdateProject",
        json={"name": name},
        headers={"Authorization": f"Bearer {token}"},
    )

    res_json = res.json()
    new_id = res_json["id"]
    config = create_databutton_config(name, new_id)
    click.secho(
        f"✅ Created project {name}",
        fg="green",
    )
    click.echo()
    styled_url = click.style(
        f"https://next.databutton.com/projects/{new_id}", fg="cyan"
    )
    click.echo(f"You can check out your project on \n\n  {styled_url}\n")
    click.secho(
        f"Type {click.style('databutton deploy', fg='cyan')} to deploy your project."
    )
    return config


@cli.command()
def docs():
    """Launches https://docs.databutton.com"""
    click.launch("https://docs.databutton.com")


@cli.command()
def logout():
    """Removes all Databutton login info"""
    login_info = get_databutton_login_info()
    if login_info is not None:
        shutil.rmtree(get_databutton_login_path(), ignore_errors=True)
        click.secho("Logged out")
        click.secho(
            f"You can always log in again with {click.style('databutton login', fg='cyan')}"
        )
    else:
        click.secho(
            f"No Databutton user found, did you mean {click.style('databutton login', fg='cyan')}?"
        )


@cli.command()
@click.option("-t", "--token", help="Print an id token", is_flag=True, default=False)
@click.option(
    "-r",
    "--refresh-token",
    default=False,
    is_flag=True,
    help="Print the refresh token."
    + " Set DATABUTTON_TOKEN to this if you want to run databutton in ci or codespaces.",
)
def whoami(token: bool, refresh_token: bool):
    """Shows the logged in user"""
    import jwt

    try:
        id_token = get_auth_token()
    except Exception:
        click.secho("No user found.")
        click.secho(f"Log in first with {click.style('databutton login', fg='cyan')}.")
        exit(1)
    decoded = jwt.decode(id_token, options={"verify_signature": False})
    click.secho("Found logged in user")
    keys_to_print = ["name", "email", "user_id"]
    for k in keys_to_print:
        click.secho(f"  {k}: {click.style(decoded[k], fg='cyan')}")
    if token:
        click.secho(f"  token: {click.style(id_token, fg='cyan')}")
    if refresh_token:
        click.secho(
            f"  refreshToken: {click.style(get_databutton_login_info().refreshToken, fg='cyan')}"
        )
    click.echo()


@cli.command()
def print_token():
    click.echo(get_auth_token())


@cli.command()
def schedule():
    """Starts and runs the scheduler"""
    click.secho("=== Starting the databutton scheduler", fg="cyan")

    Scheduler.create()


@cli.command()
@click.argument("project_id")
@click.argument("project_dir", required=False)
@click.pass_context
def download(ctx: click.Context, project_id: str, project_dir: str):
    """
    Downloads a project locally from PROJECT_ID and into PROJECT_DIR.
    This will download the last deployed code
    """
    project_dir: Path = Path(project_dir) if project_dir is not None else Path.cwd()
    if project_dir.exists():
        if len(list(project_dir.iterdir())) > 0:
            click.secho("")
            click.secho("❌ Can not download into a folder that's not empty.")
            click.secho(
                "Please select a folder that's empty (or one that doesn't exist) :-)"
            )
            exit(1)

    # Double check that the user is logged in
    try:
        get_auth_token()
    except Exception:
        click.echo("It doesn't seem like you're logged in, but we can fix that!")
        if click.confirm("Do you want to log in right now?"):
            ctx.invoke(login)
        else:
            click.echo("You need to be logged in to download this project")
            click.secho(f"Type {click.style('databutton login', fg='cyan')} to login")
            exit(1)

    click.secho("i downloading project...", fg="cyan")
    click.secho(f"i id: {project_id}", fg="cyan")
    click.secho(f"i directory: {str(project_dir)}", fg="cyan")
    download_url = f"{get_api_url(project_id)}/projects/download"
    response = requests.get(
        download_url,
        headers={"Authorization": f"Bearer {get_auth_token()}"},
        stream=True,
    )
    # Ensure that there's a zipped response
    # This can yield a 307 to login if the user does not have access
    # We can catch that by making sure that the response needs to be a tarball
    if (
        not response.ok
        or response.headers.get("content-type") != "application/tar+gzip"
    ):
        # This would be the error message from the upstream /download API
        if "next.databutton.com/login" in response.url:
            # This is what an unauthenticated api requests looks like :'(
            # We get redirected
            click.secho("Are you sure you're logged in as the correct user?")
            click.secho(
                f"Type {click.style('databutton whoami')} to see who you're logged in as."
            )
            exit(1)
        elif "detail" in response.text:
            click.secho("This is the response from the server:")
            click.secho("  " + response.json()["detail"])
        click.echo()
        click.secho(
            f"❌ Failed to get project with id {project_id}, are you sure it's the correct id?"
        )
        click.secho(
            "If you're sure it's the right id, the following url should work and have some components:"
        )
        click.secho(f"https://next.databutton.com/projects/{project_id}", fg="cyan")
        exit(1)
    with tempfile.TemporaryFile() as tmpfile:
        tmpfile.write(response.raw.read())
        tmpfile.seek(0)
        with tarfile.open(fileobj=tmpfile) as t:
            t.extractall(project_dir)
    click.secho("Finished!", fg="green")
    click.echo()
    click.secho("Finished downloading project, we recommmend that you now do")
    click.echo()
    cd_dir = (
        project_dir.relative_to(Path.cwd())
        if project_dir.is_relative_to(Path.cwd())
        else project_dir
    )
    click.secho(f"  cd {cd_dir}", fg="cyan")
    click.secho("  pip install -r requirements.txt", fg="cyan")
    click.secho("  databutton start", fg="cyan")
    click.secho("  and when you're happy with your changes,")
    click.secho("  databutton deploy", fg="cyan")
    click.echo()


@cli.group("secrets")
@click.pass_context
def secrets_group(ctx: click.Context):
    """Manage project secrets"""
    # Shared prerequisites for all secrets commands
    # _ = require_databutton_login_info()
    pass


@secrets_group.command("add")
@click.pass_context
@click.argument("name", required=True)
@click.argument("data-file", type=click.File("r"))
def secrets_add(ctx: click.Context, name: str, data_file):
    """Add secret to project.

    NB! Input is name of secret and the name of a _file_
    containing the secret value.

    The secret value is passed to the databutton api and
    stored encrypted in the cloud.
    """
    secrets_client: SecretsApiClient = get_secrets_client()

    # Read raw bytes from file
    max_len = 64 * 1024
    value: str = data_file.read(max_len + 1)
    if len(value) > max_len:
        click.secho(f"Secret value is too long (max={max_len})")
        exit(1)

    # Remove trailing newline which is usually not part of a secret.
    # Note: If this becomes a problem, add a parameter
    #       to make it opt-out but keep it as default.
    value = value.rstrip("\r\n")

    # Store secret value
    success = secrets_client.add(name, value)
    if success:
        click.echo("Successfully added secret.")
    else:
        click.echo("Failed to add secret!")


@secrets_group.command("delete")
@click.pass_context
@click.argument("name", required=True)
def secrets_delete(ctx: click.Context, name: str):
    """Delete named secret from project"""
    secrets_client: SecretsApiClient = get_secrets_client()

    config = get_databutton_config()
    confirmed = click.confirm(
        f'This will delete the secret "{name}" from project {config.uid}, are you sure?'
    )
    if not confirmed:
        click.echo("Aborted.")

    success = secrets_client.delete(name)
    if success:
        click.echo("Successfully deleted secret.")
    else:
        click.echo("Failed to delete secret!")


@secrets_group.command("get")
@click.argument("name", required=True)
@click.pass_context
def secrets_get(ctx: click.Context, name: str):
    """Show secret value."""
    secrets_client: SecretsApiClient = get_secrets_client()
    value = secrets_client.get(name)
    click.echo(value)


@secrets_group.command("list")
@click.pass_context
def secrets_list(ctx: click.Context):
    """List names of secrets in project"""
    secrets_client: SecretsApiClient = get_secrets_client()
    secrets = secrets_client.list()
    for s in secrets:
        click.echo(s)



from dataclasses import dataclass
from typing import List, Optional, Union

from dataclasses_json import dataclass_json

from .send import send


@dataclass_json
@dataclass
class Email:
    to: Union[str, List[str]]
    subject: str
    content_text: Optional[str]
    content_html: Optional[str]


def valid_email(recipient: str) -> bool:
    # Note: We could possibly use some email validation library but it's tricky
    parts = recipient.split("@")
    if len(parts) != 2:
        return False
    return bool(parts[0] and parts[1])


def validate_email_to_arg(to: Union[str, List[str]]) -> List[str]:
    if isinstance(to, str):
        to = [to]
    if not isinstance(to, (list, tuple)) and len(to) > 0:
        raise ValueError(
            "Invalid recipient, expecting 'to' to be a string or list of strings."
        )
    invalid_emails = []
    for recipient in to:
        if not valid_email(recipient):
            invalid_emails.append(recipient)
    if invalid_emails:
        raise ValueError("\n".join(["Invalid email address(es):"] + invalid_emails))
    return to


def email(
    to: Union[str, List[str]],
    subject: str,
    content_text: Optional[str] = None,
    content_html: Optional[str] = None,
):
    """Send email notification from databutton.

    At least one of the content arguments must be present.

    A link to the project will be added at the end of the email body.

    If content_text is not provided it will be generated from
    content_html for email clients without html support,
    the result may be less pretty than handcrafted text.
    """
    send(
        Email(
            to=validate_email_to_arg(to),
            subject=subject,
            content_text=content_text,
            content_html=content_html,
        )
    )



import datetime
import importlib
import inspect
import logging
import os
import sys
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, List
from uuid import uuid4

import anyio
import schedule
from anyio import create_task_group
from dataclasses_json import dataclass_json

from databutton.decorators.jobs.logger import JobRunLogger
from databutton.decorators.jobs.utils import iso_utc_timestamp
from databutton.utils import is_running_locally

if TYPE_CHECKING:
    from databutton.utils.build import ArtifactDict

logger = logging.getLogger("databutton.schedule")


@dataclass
class DatabuttonSchedule:
    # Unique identifier
    uid: str
    # The human readable name for the job
    name: str
    # The actual name of the function
    func_name: str
    # Seconds between runs (first run will be at t0 + seconds)
    seconds: float
    # The path to the .py file where the schedule is defined
    filepath: str
    # The name of the module (i.e jobs.schedule)
    module_name: str
    # Should the job be cancelled on failure
    cancel_on_failure: bool = False
    # Should run immediately after start, then be scheduled
    run_immediately: bool = True

    def to_dict(self):
        return self.to_dict()


_schedules: List[DatabuttonSchedule] = []


def repeat_every(
    seconds: float,
    name: str = None,
    cancel_on_failure: bool = False,
    run_immediately: bool = True,
):
    def wrapper(func):
        filepath = Path(inspect.getfile(func)).relative_to(Path.cwd())
        job_name = name if name is not None else func.__name__
        uid = f"{filepath}-{func.__name__}-{job_name}"
        _schedules.append(
            DatabuttonSchedule(
                uid=uid,
                name=job_name,
                func_name=func.__name__,
                seconds=seconds,
                filepath=str(filepath),
                module_name=inspect.getmodule(func).__name__,
                cancel_on_failure=cancel_on_failure,
                run_immediately=run_immediately,
            )
        )
        return func

    return wrapper


@dataclass_json
@dataclass
class JobRun:
    timestamp: datetime.datetime
    result: Any


class Job(schedule.Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.databutton_schedule: DatabuttonSchedule = None

    def from_databutton_schedule(self, databutton_schedule: DatabuttonSchedule):
        self.databutton_schedule = databutton_schedule

        mod = importlib.import_module(databutton_schedule.module_name)
        func = getattr(mod, databutton_schedule.func_name)

        self.uid = databutton_schedule.uid
        self.name = databutton_schedule.name
        return self.do(func)

    async def run(self):
        run_id = str(uuid4())
        if self.databutton_schedule is None:
            job_id = None  # Raising instead because we need a job id below here!
        else:
            job_id = self.databutton_schedule.uid

        if job_id is None and not is_running_locally():
            raise Exception(
                "Running job without a databutton schedule, need a job id to continue"
            )

        if self._is_overdue(datetime.datetime.now()):
            logger.debug("Cancelling job %s", self)
            return schedule.CancelJob

        with JobRunLogger(logger, job_id=job_id, run_id=run_id) as jobrun:
            try:
                if inspect.iscoroutinefunction(self.job_func):
                    ret = await self.job_func()
                else:
                    ret = self.job_func()
                jobrun.success = True
            except Exception as e:
                jobrun.success = False

                # This needs to be printed, as the stdout is handled by JobRunner
                # Otherwise we would need to configure the thing to do standard out
                print(traceback.format_exc())

                if os.environ.get("SENTRY_DSN"):
                    import sentry_sdk

                    with sentry_sdk.push_scope() as scope:
                        scope.set_tag("job.uid", job_id)
                        scope.set_tag("job.run_id", run_id)
                        # Not bothering with fixing this now but I think these are wrong:
                        # This will be when the previous run started?
                        scope.set_extra("job.last_run", self.last_run)
                        # This will be when this run was supposed to start?
                        scope.set_extra("job.next_run", self.next_run)
                        sentry_sdk.capture_exception(e)

                if self.databutton_schedule.cancel_on_failure:
                    ret = schedule.CancelJob
                else:
                    ret = e  # Or None? Depends on how we use the jobrun history later

            self.last_run = jobrun.start_time

            self._schedule_next_run()

            if self._is_overdue(datetime.datetime.now()):
                logger.debug("Cancelling job %s", self)
                return schedule.CancelJob

            # Used on jobrun exit, not set if the cancel above hits
            # (but I'm not sure I follow the original logic here)
            jobrun.next_run_time = iso_utc_timestamp(self.next_run)

        return ret


class Scheduler(schedule.Scheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jobs: List[Job] = []
        self.running_process = None

    def load_schedules(self, components: "ArtifactDict"):
        should_run_immediately: List[Job] = []
        for sched in components.schedules:
            job = self.every(sched.seconds).seconds.from_databutton_schedule(sched)

            if sched.run_immediately:
                should_run_immediately.append(job)

        async def run_continously():
            await self.run_pending(override=should_run_immediately)
            while True:
                await self.run_pending()
                await anyio.sleep(1)

        return run_continously()

    # Have a proxy so we can have a clean coroutine
    async def _create(self, components):
        await self.load_schedules(components)

    @classmethod
    def create(cls):
        from databutton.utils.build import read_artifacts_json

        sys.path.append(".")

        sched = Scheduler()
        components = read_artifacts_json()
        try:
            anyio.run(sched._create, components, backend="trio")
        except KeyboardInterrupt as e:
            # Exit gracefully it interrupted
            logger.debug(e)
            sys.exit(0)
        return sched

    async def run_pending(self, override: List[Job] = None):
        job_pool = override if override else self.jobs
        force = True if override else False
        runnable_jobs = [job for job in job_pool if job.should_run or force]
        async with create_task_group() as tg:
            for job in sorted(runnable_jobs):
                tg.start_soon(job.run)

    def every(self, interval=1):
        job = Job(interval, self)
        return job



import os
from typing import Any

import requests

from databutton.utils import get_api_url, get_auth_token, get_databutton_config


def get_api_headers() -> dict[str]:
    return {
        "Authorization": "Bearer " + get_auth_token(),
        "x-databutton-release": os.environ.get("DATABUTTON_RELEASE"),
    }


def send(message: Any):
    """Internal helper to make request towards notification api."""

    # The idea here is that perhaps we can reuse some notification
    # queue system etc across kinds, e.g. email, slack, etc.
    kind = message.__class__.__name__.lower()

    project_id = get_databutton_config().uid
    url = f"{get_api_url(project_id)}/notify/{kind}"

    headers = get_api_headers()

    # TODO: Use pydantic.BaseModel etc instead of dataclasses_json to avoid .to_dict()?
    body = {"message": message.to_dict()}

    res = requests.post(url, headers=headers, json=body)
    if not res.ok:
        raise Exception(f"Failed to send {kind} notification")
    return



import logging
import os
import subprocess
import tarfile
from fnmatch import fnmatch
from pathlib import Path
from time import sleep
from typing import IO, Generator

import requests

from databutton.utils import ProjectConfig, get_auth_token, get_databutton_login_info


def get_firebase(path: str):
    token = get_auth_token()
    return requests.get(
        f"https://firestore.googleapis.com/v1/projects/databutton/databases/(default)/documents/{path}",
        headers={"Authorization": f"Bearer {token}"},
    )


def push_firebase(path: str):
    pass


def wait_for_firebase_existence(path, retries=10):
    res = get_firebase(path)
    if retries == 0:
        return None
    if res.status_code == 404:
        # This doesn't exist (yet), try again in 2 seconds
        sleep(2)
        return wait_for_firebase_existence(path, retries=retries - 1)
    return res


def get_deployment(deployment_id):
    return wait_for_firebase_existence(f"deployments/{deployment_id}").json()


def get_build_status(build_id: str) -> str:
    json = get_firebase(f"cloud-builds/{build_id}").json()
    return json["fields"]["status"]["stringValue"]


def get_build_id_from_deployment(deployment_id: str) -> str:
    deployment = get_deployment(deployment_id)
    build_id = deployment["fields"]["buildId"]["stringValue"]
    return build_id


def listen_to_build(deployment_id: str) -> Generator[str, None, None]:
    build_id = get_build_id_from_deployment(deployment_id)
    status = get_build_status(build_id)

    prev = status
    yield status
    while status not in ["SUCCESS", "FAILURE", "CANCELLED"]:
        sleep(5)
        status = get_build_status(build_id)
        if status != prev:
            yield status
            prev = status
    yield status


def create_archive(
    tmpfile: IO[bytes], source_dir: Path = Path.cwd(), config: ProjectConfig = None
) -> str:
    ignore_files = config.exclude
    if ".databutton" in ignore_files:
        # We need this for deploying.
        ignore_files.remove(".databutton")
    generated_requirements = False
    if os.path.exists("pyproject.toml") and not os.path.exists("requirements.txt"):
        # This is a poetry project, export requirements.txt from poetry
        subprocess.run(
            "poetry export -f requirements.txt --output requirements.txt --without-hashes",
            shell=True,
        )
        generated_requirements = True

    def exclude_filter(tarinfo: tarfile.TarInfo):
        for ignore in ignore_files:
            if fnmatch(tarinfo.name, ignore):
                return None
        return tarinfo

    with tarfile.open(fileobj=tmpfile, mode="w:gz") as tar:
        for fn in source_dir.iterdir():
            tar.add(fn, arcname=fn.relative_to(source_dir), filter=exclude_filter)
    if generated_requirements:
        os.remove("requirements.txt")

    # Reset the reader so we can get the contents
    tmpfile.flush()
    tmpfile.seek(0)

    return tmpfile


def get_signed_url_for_upload(
    project_id: str, deployment_id: str, databutton_token: str, auth_token: str
):
    get_signed_url = (
        "https://europe-west1-databutton.cloudfunctions.net/upload_project_signer"
    )
    res = requests.post(
        get_signed_url,
        headers={"Authorization": f"Bearer {auth_token}"},
        json={
            "project_id": project_id,
            "deployment_id": deployment_id,
            "databutton_token": databutton_token,
        },
    )
    if not res.ok:
        raise Exception("Could not upload")
    return res.json()["signed_url"]


def upload_archive(config: ProjectConfig, deployment_id: str, tmpfile: IO[bytes]):
    refresh_token = get_databutton_login_info().refreshToken
    signed_url = get_signed_url_for_upload(
        project_id=config.uid,
        deployment_id=deployment_id,
        databutton_token=refresh_token,
        auth_token=get_auth_token(),
    )
    signed_headers = {
        "Authorization": f"Bearer {get_auth_token()}",
        "x-goog-meta-databutton-token": refresh_token,
        "content-type": "application/tar+gzip",
    }

    response = requests.put(signed_url, headers=signed_headers, data=tmpfile)
    if not response.ok:
        logging.error(response.json())
        raise Exception("Could not upload archive")
    return tmpfile



import streamlit as st
from htbuilder import HtmlElement, a, div, hr, img, p, styles
from htbuilder.units import percent, px


def image(src_as_string, **style):
    return img(src=src_as_string, style=styles(**style))


def link(link, text, **style):
    return a(_href=link, _target="_blank", style=styles(**style))(text)


def layout(*args):

    style = """
    <style>
      # MainMenu {visibility: hidden;}
      footer {visibility: hidden;}
     .stApp { bottom: 50px; }
    </style>
    """

    style_div = styles(
        position="fixed",
        left=0,
        bottom=0,
        margin=px(0, 0, 0, 0),
        width=percent(100),
        color="white",
        text_align="center",
        height="auto",
        opacity=1,
    )

    style_hr = styles(
        display="block",
        margin=px(8, 8, "auto", "auto"),
        border_style="inset",
        border_width=0,
    )

    body = p()
    foot = div(style=style_div)(hr(style=style_hr), body)

    st.markdown(style, unsafe_allow_html=True)

    for arg in args:
        if isinstance(arg, str):
            body(arg)

        elif isinstance(arg, HtmlElement):
            body(arg)

    st.markdown(str(foot), unsafe_allow_html=True)


def footer():
    stst = (
        "<a href='http://www.databutton.com'>"
        + str(
            image(
                "https://uploads-ssl.webflow.com/6213e8906bdf786dcccbce62/621d20f1031cff7b094664f2_large_fav.png",
                width=px(25),
                height=px(25),
            )
        )
        + "</a>"
    )
    myargs = [stst, "&nbsp Made with <b>Databutton</b> ", " "]
    layout(*myargs)



import json
from pathlib import Path

from databutton_web import get_static_file_path
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel
from starlette.responses import FileResponse

from databutton.utils import get_databutton_login_path

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


class AuthSuccess(BaseModel):
    refreshToken: str
    uid: str


@app.get("/")
def index():
    return FileResponse(f"{get_static_file_path()}/login.html")


@app.post("/")
async def auth_success(auth: AuthSuccess):
    filename = Path(get_databutton_login_path(), f"{auth.uid}.json")
    # Create files with necessary folder structure if it does not exist
    filename.parent.mkdir(exist_ok=True, parents=True)
    with open(filename, "w") as f:
        f.write(
            json.dumps({"uid": auth.uid, "refreshToken": auth.refreshToken}, indent=2)
        )

    return {"success": True}


app_dir = get_static_file_path()
app.mount("/", StaticFiles(directory=app_dir), name="app")



import logging
import os

import sentry_sdk
from databutton_web import get_static_file_path
from fastapi import FastAPI, Response, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

from databutton.decorators.apps.streamlit import create_streamlit_router
from databutton.utils import get_databutton_config, get_databutton_login_info
from databutton.utils.build import read_artifacts_json
from databutton.utils.log_status import log_devserver_screen
from databutton.version import __version__

logger = logging.getLogger("databutton.webserver")
logging.basicConfig(
    format="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
)
config = get_databutton_config()
components = read_artifacts_json()

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Set up proxies for streamlit apps
# They are ran as containers somewhere
for st in components.streamlit_apps:
    app.include_router(create_streamlit_router(st))


@app.get("/")
async def index():
    return RedirectResponse(url="/index.html")


@app.on_event("startup")
async def startup():
    log_devserver_screen(components=components)


app_dir = get_static_file_path()
app.mount("/static", StaticFiles(directory=".databutton"), name=".databutton")
app.mount("/", StaticFiles(directory=app_dir), name="app")


@app.get("/healthz")
async def healthz():
    return Response(status_code=status.HTTP_200_OK)


if os.environ.get("SENTRY_DSN"):
    logging.info("Found SENTRY_DSN, logging errors")
    sentry_sdk.init(dsn=os.environ.get("SENTRY_DSN"))
    config = get_databutton_config()
    login_info = get_databutton_login_info()
    sentry_tags = {
        "databutton_release": os.environ.get("DATABUTTON_RELEASE", "latest"),
        "databutton_project_id": config.uid,
        "databutton_project_name": config.name,
        "databutton_version": __version__,
        "databutton_user_id": login_info.uid if login_info else None,
    }
    sentry_sdk.set_user({"uid": login_info.uid if login_info else None})
    for k, v in sentry_tags.items():
        sentry_sdk.set_tag(k, v)

    app = SentryAsgiMiddleware(app)



import importlib
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from dataclasses_json import dataclass_json

import databutton as db
from databutton.decorators.apps.streamlit import StreamlitApp
from databutton.decorators.jobs.schedule import DatabuttonSchedule
from databutton.helpers import parse
from databutton.kubernetes.generate import generate_manifest
from databutton.utils import get_databutton_components_path
from databutton.version import __version__


@dataclass_json
@dataclass
class ArtifactDict:
    streamlit_apps: List[StreamlitApp] = field(default_factory=List)
    schedules: List[DatabuttonSchedule] = field(default_factory=List)

    def __eq__(self, other: "ArtifactDict"):
        if isinstance(other, ArtifactDict):
            return self.to_json() == other.to_json()
        return False


def generate_artifacts_json():
    # Sort the apps so that the port proxy remains stable
    for i, st in enumerate(sorted(db.apps._streamlit_apps, key=lambda x: x.route)):
        st.port = 8501 + i
    artifacts = ArtifactDict(
        streamlit_apps=[st for st in db.apps._streamlit_apps],
        schedules=[sched for sched in db.jobs._schedules],
    )
    return artifacts


def write_artifacts_json(artifacts: ArtifactDict):
    if get_databutton_components_path().exists():
        existing = read_artifacts_json()
        if existing == artifacts:
            return
    with open(get_databutton_components_path(), "w") as f:
        f.write(artifacts.to_json(indent=2))


def read_artifacts_json() -> ArtifactDict:
    with open(get_databutton_components_path(), "r") as f:
        return ArtifactDict.from_json(f.read())


def generate_components(rootdir: Path = Path.cwd()):
    normalized_rootdir = rootdir.resolve().relative_to(Path.cwd())
    if rootdir.resolve() not in sys.path:
        sys.path.insert(0, str(rootdir.resolve()))
    # Find all directive modules and import them
    imports = parse.find_databutton_directive_modules(rootdir=normalized_rootdir)

    # Clean the existing artifacts, generate new one
    # TODO: Have a cache mechanism to improve performance
    # shutil.rmtree(Path(".databutton"), ignore_errors=True)
    Path(".databutton").mkdir(exist_ok=True)
    decorator_modules = {}
    for name in imports:
        mod = importlib.import_module(name)
        decorator_modules[name] = mod
    # Write the artifacts
    # Sort the apps so that the port proxy remains stable
    artifacts = generate_artifacts_json()
    write_artifacts_json(artifacts)

    # Copy the Dockerfile
    parent_folder = Path(__file__).parent.parent
    current_dockerfile_path = Path(parent_folder, "docker", "Dockerfile")

    docker_folder_path = Path(".databutton", "docker")
    docker_folder_path.mkdir(exist_ok=True, parents=True)
    dest_dockerfile_path = Path(docker_folder_path, "Dockerfile")
    with open(current_dockerfile_path, "r") as original:
        contents = original.read()
        with open(dest_dockerfile_path, "w") as dest:
            dest.write(
                # Overwrite image
                contents.replace("REPLACE_ME_VERSION", __version__)
            )

    # Generate a kubernetes manifest
    generate_manifest(artifacts)

    # Find apps that are not used and nuke the remainding time
    app_dir = Path(".databutton", "app")
    if app_dir.exists():
        existing_app_files = list(app_dir.iterdir())
        new_fpaths = [app.filename for app in artifacts.streamlit_apps]
        for f in existing_app_files:
            if str(f) not in new_fpaths:
                f.unlink()

    return artifacts



import base64


def str_to_base64str(value: str) -> str:
    return base64.urlsafe_b64encode(value.encode("utf8")).decode("utf8")


def base64str_to_str(value_base64: str) -> str:
    return base64.urlsafe_b64decode(value_base64).decode("utf8")



import asyncio
import inspect
import logging
from dataclasses import dataclass
from hashlib import md5
from pathlib import Path
from typing import List, Optional
from uuid import UUID

import httpx
from fastapi import APIRouter, Request, Response, WebSocket, status
from httpx import AsyncClient
from starlette.background import BackgroundTask
from starlette.responses import StreamingResponse
from websockets import WebSocketClientProtocol, connect

from databutton.utils.config import MAX_WEBSOCKET_MESSAGE_SIZE_IN_MB

# Decorator for defining a method as a streamlit app


DEFAULT_MEMORY = "256Mi"
DEFAULT_CPU = "0.2"


@dataclass
class StreamlitApp:
    filename: str
    uid: str
    name: str
    route: str
    memory: str = DEFAULT_MEMORY
    cpu: str = DEFAULT_CPU
    port: int = 0


# Global var to store all streamlit apps
_streamlit_apps: List[StreamlitApp] = []


def streamlit(
    route: str,
    name: str = None,
    memory: Optional[str] = None,
    cpu: Optional[str] = None,
):
    def app(func):
        cleaned_route = route if route.endswith("/") else route + "/"
        hasher = md5()
        hasher.update(cleaned_route.encode("utf-8"))
        uid = str(UUID(hasher.hexdigest()))
        filename = Path(".databutton", "app", f"tmp-{uid}.py")
        filename.parent.mkdir(parents=True, exist_ok=True)

        module_name = inspect.getmodule(func).__name__
        func_name = func.__name__
        import_statement = f"from {module_name} import {func_name}"

        with open(filename, "w") as f:
            f.write(import_statement)
            f.write("\n")
            f.write("\n")
            f.write(f"{func_name}()")
        st = StreamlitApp(
            filename=str(filename),
            route=cleaned_route,
            name=name if name else func_name,
            uid=uid,
        )
        if memory:
            st.memory = memory
        if cpu:
            st.cpu = cpu
        _streamlit_apps.append(st)
        return func

    return app


def create_streamlit_router(app: StreamlitApp):

    router = APIRouter()

    @router.websocket(app.route + "stream")
    async def handle_proxied_websocket(ws_client: WebSocket):
        try:
            await ws_client.accept()
            port = app.port
            if port is None:
                return Response(status_code=status.HTTP_404_NOT_FOUND)
            max_size = MAX_WEBSOCKET_MESSAGE_SIZE_IN_MB * int(1e6)
            async with connect(
                f"ws://localhost:{port}/stream", max_size=max_size
            ) as ws_server:
                fwd_task = asyncio.create_task(forward(ws_client, ws_server))
                rev_task = asyncio.create_task(reverse(ws_client, ws_server))
                await asyncio.gather(fwd_task, rev_task)
        except Exception as e:
            # Simply ignore messages. It's not this one's job to make sure it's up and running
            logging.debug("Error in websocket proxy", extra=e)

    @router.api_route(
        app.route + "{rest:path}",
        methods=["GET", "POST", "PATCH", "PUT"],
    )
    async def _reverse_proxy(request: Request, rest: str):
        # Find correct app
        url = httpx.URL(path=rest, query=request.url.query.encode("utf-8"))
        client = AsyncClient(base_url=f"http://localhost:{app.port}/")
        req = client.build_request(
            request.method,
            url,
            headers=request.headers.raw,
            content=await request.body(),
        )
        r = await client.send(req, stream=True)

        return StreamingResponse(
            r.aiter_raw(),
            status_code=r.status_code,
            background=BackgroundTask(r.aclose),
            headers=r.headers,
        )

    return router


async def forward(ws_client: WebSocket, ws_server: WebSocketClientProtocol):
    while True:
        data = await ws_client.receive_bytes()
        await ws_server.send(data)


async def reverse(ws_client: WebSocket, ws_server: WebSocketClientProtocol):
    while True:
        data = await ws_server.recv()
        await ws_client.send_text(data)



"""
Create data apps

.. currentmodule:: databutton
.. moduleauthor:: Databutton <support@databutton.com>
"""

from . import notify  # noqa
from . import secrets  # noqa
from . import storage  # noqa
from .branding import streamlit  # noqa
from .decorators import apps  # noqa
from .decorators import jobs  # noqa
from .version import __version__  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['databutton',
 'databutton.auth',
 'databutton.branding',
 'databutton.branding.streamlit',
 'databutton.dataframes',
 'databutton.debug',
 'databutton.decorators',
 'databutton.decorators.apps',
 'databutton.decorators.jobs',
 'databutton.helpers',
 'databutton.kubernetes',
 'databutton.notify',
 'databutton.project',
 'databutton.secrets',
 'databutton.server',
 'databutton.storage',
 'databutton.utils']

package_data = \
{'': ['*'], 'databutton': ['docker/*']}

install_requires = \
['PyJWT>=2.4.0,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'alive-progress>=2.4.1,<3.0.0',
 'anyio[trio]>=3.6.1,<4.0.0',
 'click>=7.0,<8.1',
 'databutton-web>=0.16.0,<0.17.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'fastapi>=0.81.0,<0.82.0',
 'htbuilder>=0.6.0,<0.7.0',
 'httpx>=0.23.0,<0.24.0',
 'pandas>=0.21.0',
 'psutil>=5.9.1,<6.0.0',
 'schedule>=1.1.0,<2.0.0',
 'sentry-sdk==1.9.7',
 'streamlit>=1.9.0,<2.0.0',
 'uvicorn>=0.18.0,<0.19.0',
 'watchfiles>=0.16.0,<0.17.0',
 'watchgod>=0.8.2,<0.9.0',
 'websockets>=10.3,<11.0']

entry_points = \
{'console_scripts': ['databutton = databutton.cli:cli']}

setup_kwargs = {
    'name': 'databutton',
    'version': '0.23.0',
    'description': 'The CLI for databutton.com',
    'long_description': "# databutton-cli\n[![semantic-release: angular](https://img.shields.io/badge/semantic--release-angular-e10079?logo=semantic-release)](https://github.com/semantic-release/semantic-release)\n[![PyPI version fury.io](https://badge.fury.io/py/databutton.svg)](https://pypi.python.org/pypi/databutton/)\n[![PyPI download week](https://img.shields.io/pypi/dw/databutton.svg)](https://pypi.python.org/pypi/databutton/)\n![release](https://github.com/databutton/databutton-cli/actions/workflows/release.yaml/badge.svg)\n\n\n\nThe CLI for building and deploying databutton projects\n\n## Getting Started\n\n```bash\nUsage: databutton [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -v, --verbose  Enable verbose logging\n  --help         Show this message and exit.\n\nCommands:\n  build    Build the project, built components will be found in .databutton\n  create   Create a Databutton project in the provided project-directory\n  deploy   Deploy your project to Databutton\n  docs     Launches https://docs.databutton.com\n  init     Creates a new project in Databutton and writes to databutton.json\n  login    Login to Databutton\n  logout   Removes all Databutton login info\n  serve    Starts a web server for production.\n  start    Run the Databutton development server\n  version  Get the library version.\n  whoami   Shows the logged in user\n```\n\n## Developing\n\n### Prerequisites\nThis project uses poetry, so if you haven't already;\n\n`pip install poetry`\n\n### Install dependencies\n\n`poetry install`\n\n### Test\n\n`poetry run pytest -s`\n\n### Lint\n`poetry run flake8 .`\n`poetry run black .`\n`poetry run isort .`\n\nAll these are being run in a github action on pull requests and the main branch.\n\n### Test locally in another package\n\nTo test in another package, you can simply\n\n`pip install -e .` assuming you're in this folder. If not, replace the `.` with the path to the `databutton-cli` folder.\n\n## Authors\n\n* **Databutton** - *Initial work* - [github](https://github.com/databutton)\n\n## License: Copyright (c) Databutton\n\nAll rights reserved.\n\n\n",
    'author': 'Databutton',
    'author_email': 'hi@databutton.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import json
import logging
import os
import pathlib
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import List, Optional, Union

import requests
from packaging.version import Version, parse

from databutton.version import __version__

logger = logging.getLogger("databutton.utils")


DEFAULT_GLOB_EXCLUDE = ["venv", ".venv", "__pycache__", ".databutton", ".git"]


@dataclass
class ProjectConfig:
    uid: str
    name: str
    # List of fnmatch patterns to exclude, similar to .gitignore
    exclude: Optional[List[str]] = field(default_factory=lambda: DEFAULT_GLOB_EXCLUDE)


CONFIG_FILENAME = "databutton.json"


def get_databutton_config_path() -> Path:
    project_directory = pathlib.Path.cwd()
    retries = 1  # Increase to re-enable feature looking in parent directories or clean up code
    for _ in range(retries):
        filepath = project_directory / CONFIG_FILENAME
        if filepath.exists():
            return filepath
        project_directory = project_directory.parent
        if not project_directory.is_relative_to(pathlib.Path.home()):
            break
    raise FileNotFoundError(
        "Could not find databutton.json config file in parent directories."
    )


def get_project_directory() -> Path:
    return get_databutton_config_path().parent


def get_databutton_config() -> ProjectConfig:
    envvar = os.environ.get("DATABUTTON_PROJECT_ID")
    if envvar is not None:
        return ProjectConfig(name="env", uid=envvar)
    with open(get_databutton_config_path(), "r") as f:
        config = json.load(f)
        return ProjectConfig(
            name=config["name"],
            uid=config["uid"],
            exclude=config["exclude"],
        )


_cached_databutton_config: Optional[ProjectConfig] = None


def get_cached_databutton_config() -> Optional[ProjectConfig]:
    global _cached_databutton_config
    if _cached_databutton_config is None:
        _cached_databutton_config = get_databutton_config()
    return _cached_databutton_config


def create_databutton_config(
    name: str, uid: str, project_directory: Path = None
) -> ProjectConfig:
    if project_directory is None:
        project_directory = pathlib.Path.cwd()
    config = ProjectConfig(
        name=name,
        uid=uid,
        exclude=DEFAULT_GLOB_EXCLUDE,
    )
    with open(project_directory / CONFIG_FILENAME, "w") as f:
        f.write(json.dumps(config.__dict__, indent=2))
        return config


def get_api_url(uid: str) -> str:
    u = os.environ.get("DATABUTTON_API_URL")
    return u or f"https://p{uid}.dbtn.app/dbtn"


@dataclass
class LoginData:
    refreshToken: str
    uid: str


def get_databutton_login_info() -> Optional[LoginData]:
    if "DATABUTTON_TOKEN" in os.environ:
        return LoginData(refreshToken=os.environ["DATABUTTON_TOKEN"], uid="token")

    auth_path = get_databutton_login_path()
    auth_path.mkdir(exist_ok=True, parents=True)

    uids = [f for f in os.listdir(auth_path) if f.endswith(".json")]
    if len(uids) > 0:
        # Just take a random one for now
        with open(auth_path / uids[0]) as f:
            config = json.load(f)
            return LoginData(uid=config["uid"], refreshToken=config["refreshToken"])
    return None


def get_databutton_login_path():
    return Path(
        os.environ.get(
            "DATABUTTON_LOGIN_PATH", Path(Path.home(), ".config", "databutton")
        )
    )


def get_databutton_components_path():
    return Path(".databutton", "artifacts.json")


FIREBASE_CONFIG = {
    "apiKey": "AIzaSyAdgR9BGfQrV2fzndXZLZYgiRtpydlq8ug",
    "authDomain": "databutton.firebaseapp.com",
    "projectId": "databutton",
    "storageBucket": "databutton.appspot.com",
    "databaseURL": "",
}
storage_bucket = (
    f"https://firebasestorage.googleapis.com/v0/b/{FIREBASE_CONFIG['storageBucket']}"
)


def upload_to_bucket(
    file_buf, config: ProjectConfig, key: str, content_type: str = "text/csv"
):
    url = get_signed_dataframe_url(project_id=config.uid, key=key, method="PUT")

    response = requests.put(url, data=file_buf, headers={"content-type": content_type})
    if not response.ok:
        raise Exception(f"Could not upload to path {key}")
    return response


def get_signed_dataframe_url(project_id: Optional[str], key: str, method: str) -> str:
    """
    Gets a signed url to either get or put a dataframes
    args:
        key: The dataframe key
        method: GET or PUT
    """
    project_id = project_id or get_databutton_config().uid
    res = requests.get(
        "https://europe-west1-databutton.cloudfunctions.net/dataframes_signer",
        params={"data_key": key, "method": method, "project_id": project_id},
        headers={
            "Authorization": f"Bearer {get_auth_token()}",
            "x-databutton-release": os.environ.get("DATABUTTON_RELEASE"),
        },
    )
    if not res.ok:
        logger.debug(f"Failed getting dataframe {res.text}")
        raise Exception(f"Failed fetching dataframe with status_code {res.status_code}")
    try:
        return res.json()["signed_url"]
    except Exception as e:
        logger.debug("Failed getting dataframe")
        raise e


def download_from_bucket(key: str, config: ProjectConfig):
    url = get_signed_dataframe_url(config.uid, key, method="GET")

    response = requests.get(url)
    if not response.ok:
        if response.status_code == 404:
            raise FileNotFoundError(f"Could not find {key}")
        raise Exception(f"Could not download {key}")
    return response


_cached_auth_token = None


def get_auth_token() -> str:
    global _cached_auth_token
    # This has a 15 minute cache
    if _cached_auth_token is not None and time() - _cached_auth_token[0] > 60 * 15:
        _cached_auth_token = None
    if _cached_auth_token is None:
        login_info = get_databutton_login_info()
        if login_info is None:
            raise Exception(
                "Could not find any login information."
                "\nAre you sure you are logged in?"
            )
        res = requests.post(
            f"https://securetoken.googleapis.com/v1/token?key={FIREBASE_CONFIG['apiKey']}",
            {"grant_type": "refresh_token", "refresh_token": login_info.refreshToken},
        )
        if not res.ok:
            raise Exception("Could not authenticate")
        json = res.json()
        _cached_auth_token = (time(), json["id_token"])
    return _cached_auth_token[1]


def create_databutton_cloud_project(name: str):
    """Creates a Databutton Cloud Project"""
    token = get_auth_token()

    res = requests.post(
        "https://europe-west1-databutton.cloudfunctions.net/createOrUpdateProject",
        json={"name": name},
        headers={"Authorization": f"Bearer {token}"},
    )

    res_json = res.json()
    new_id = res_json["id"]
    return new_id


def get_build_logs(build_id: str) -> str:
    log_url_response = requests.get(
        "https://europe-west1-databutton.cloudfunctions.net/get_cloud_build_logs",
        params={"build_id": build_id},
        headers={"Authorization": f"Bearer {get_auth_token()}"},
    )
    log_url = log_url_response.json()["signed_url"]
    return log_url


def get_newest_pip_databutton_version() -> Version:
    res = requests.get("https://pypi.python.org/pypi/databutton/json").json()
    v = parse("0")
    for version in res.get("releases").keys():
        vv = parse(version)
        if not vv.is_prerelease:
            v = max(v, vv)
    return v


def new_databutton_version_exists() -> Union[Version, bool]:
    try:
        current = parse(__version__)
        newest = get_newest_pip_databutton_version()
        if newest > current:
            return newest
        return False
    except Exception as e:
        logger.debug("Could not fetch new version", e)


def current_databutton_release():
    return os.environ.get("DATABUTTON_RELEASE", "")


def is_running_locally():
    return current_databutton_release() == ""



import io

import pandas as pd

from databutton.utils import (
    download_from_bucket,
    get_databutton_config,
    upload_to_bucket,
)


def get(key: str, ignore_not_found=True) -> pd.DataFrame:
    config = get_databutton_config()
    try:
        res = download_from_bucket(key, config)
        return pd.read_feather(io.BytesIO(res.content))
    except FileNotFoundError as e:
        if ignore_not_found:
            return pd.DataFrame()
        raise e


def put(df: pd.DataFrame, key: str):
    config = get_databutton_config()
    buf = io.BytesIO()
    df.to_feather(buf)
    # Reset to be able to upload
    buf.seek(0)

    upload_to_bucket(buf, config, key, content_type="vnd.apache.arrow.file")

    return True


def concat(
    key: str,
    other: pd.DataFrame,
    ignore_index: bool = False,
    verify_integrity: bool = False,
    sort: bool = False,
):
    df = get(key)
    new_df = pd.concat(
        [df, other],
        ignore_index=ignore_index,
        verify_integrity=verify_integrity,
        sort=sort,
    )
    put(new_df, key=key)
    return new_df


def add(key: str, entry):
    return concat(key=key, other=pd.DataFrame(entry, index=[0]), ignore_index=True)


def clear(key: str):
    """Empty the data at a certain key, leaving you with an empty dataframe on the next .get"""
    return put(pd.DataFrame(data=None).reset_index(), key=key)



import os

import requests

from databutton.utils import get_api_url, get_auth_token, get_databutton_config

from .utils import base64str_to_str, str_to_base64str


class SecretsApiClient:
    def __init__(self, project_id: str = None):
        if project_id is None:
            project_id = get_databutton_config().uid
        self.base_url = get_api_url(project_id) + "/secrets"
        self.project_id = project_id

    def _headers(self):
        token = get_auth_token()
        if token is None or token == "":
            raise EnvironmentError(
                "Missing auth token. Are you logged in to databutton?"
            )
        return {
            "Authorization": f"Bearer {token}",
            "x-databutton-release": os.environ.get("DATABUTTON_RELEASE"),
        }

    def add(self, name: str, value: str) -> bool:
        # Note: Api expects base64 encoding of the raw bytes.
        res = requests.post(
            self.base_url + "/add",
            headers=self._headers(),
            json={"name": name, "value": str_to_base64str(value)},
        )
        if not res.ok:
            raise Exception("Failed to add secret")
        body = res.json()
        return body["name"] == name and body["added"] is True

    def delete(self, name: str) -> bool:
        res = requests.post(
            self.base_url + "/delete",
            headers=self._headers(),
            json={"name": name},
        )
        if not res.ok:
            raise Exception("Failed to delete secret")
        body = res.json()
        return body["name"] == name and body["deleted"] is True

    def _get_secret_data(self, name: str) -> dict:
        """Get secret value and metadata.

        Note: Returning value as field in json object such that we
        can easily include more metadata in the response if we want.
        This means the secret bytes must be formatted as a valid string,
        so we're using base64 for the transfer.
        """
        # Using post to pass parameters in body
        res = requests.post(
            self.base_url + "/get",
            headers=self._headers(),
            json={"name": name},
        )
        if not res.ok:
            raise Exception("Failed to read secret")
        return res.json()

    def get(self, name: str) -> str:
        value_base64 = self._get_secret_data(name).get("value")
        return base64str_to_str(value_base64)

    def list(self) -> list[str]:
        res = requests.get(
            self.base_url + "/list",
            headers=self._headers(),
        )
        if not res.ok:
            raise Exception("Failed to list secrets")
        body = res.json()
        return sorted([s["name"] for s in body["secrets"]])



import click

from databutton import __version__
from databutton.utils.build import ArtifactDict


def log_devserver_screen(components: ArtifactDict):
    click.echo()
    local_url = click.style("http://localhost:8000", fg="cyan")
    click.echo(
        click.style(
            f"databutton v{__version__} dev server running on {local_url}",
            fg="green",
        )
    )

    if len(components.streamlit_apps) > 0:
        click.echo()
        click.echo(click.style("apps", fg="cyan", bold=True))
        for st in components.streamlit_apps:
            click.echo(f"name: {st.name:<30} > http://localhost:8000{st.route}")

    if len(components.schedules) > 0:
        click.echo()
        click.echo(click.style("jobs", fg="cyan", bold=True))
        for job in components.schedules:
            click.echo(f"name: {job.name:<30}> every: {job.seconds}s")

    click.echo()
    click.echo(click.style("listening to changes...", fg="cyan"))
    click.echo()



import logging
from contextlib import redirect_stdout

from databutton.decorators.jobs.utils import (
    create_job_run,
    iso_utc_timestamp_now,
    push_job_log,
    update_job_run,
)
from databutton.utils import is_running_locally

logger = logging.getLogger("databutton.logpusher")


class JobRunLogger:
    def __init__(
        self,
        logger: logging.Logger,
        job_id: str,
        run_id: str,
        level: int = logging.INFO,
    ):
        self.logger = logger
        self.level = level
        self.job_id = job_id
        self.run_id = run_id
        self.success = False
        self.start_time = None
        self.end_time = None
        self.next_run_time = None
        self._redirector = redirect_stdout(self)

    def write(self, msg: str):
        # from https://johnpaton.net/posts/redirect-logging/
        if msg and not msg.isspace():
            stripped = msg.strip()
            if not is_running_locally():
                # deployed version
                try:
                    push_job_log(
                        run_id=self.run_id,
                        msg=stripped,
                    )
                except Exception:
                    import traceback

                    logger.error(traceback.format_exc())
            self.logger.log(self.level, stripped)

    def __enter__(self):
        self._redirector.__enter__()
        self.start_time = iso_utc_timestamp_now()
        if not is_running_locally():
            create_job_run(self.job_id, self.run_id, self.start_time)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_time = iso_utc_timestamp_now()
        if not is_running_locally():
            update_job_run(
                self.job_id,
                self.run_id,
                self.start_time,
                self.end_time,
                self.next_run_time,
                self.success,
            )
        # let contextlib do any exception handling here
        self._redirector.__exit__(exc_type, exc_value, traceback)

    def flush(self):
        pass



import asyncio
import hashlib
import logging
import os
import shutil
import signal
import traceback
from asyncio.subprocess import Process
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import anyio
import click
import psutil
from anyio import CancelScope, create_task_group, open_process, open_signal_receiver
from anyio.streams.buffered import BufferedByteReceiveStream
from watchfiles import Change, DefaultFilter, arun_process, awatch

from databutton.decorators.apps.streamlit import StreamlitApp
from databutton.decorators.jobs.schedule import Scheduler
from databutton.utils import get_databutton_components_path
from databutton.utils.build import generate_components, read_artifacts_json
from databutton.version import __version__

logger = logging.getLogger("databutton.start")
awatch_logger = logging.getLogger("watchfiles.main")
awatch_logger.setLevel(logging.CRITICAL)


class DatabuttonFilter(DefaultFilter):
    def __init__(
        self,
        *,
        ignore_paths: Optional[Sequence[Union[str, Path]]] = None,
        extra_extensions: Sequence[str] = (),
        include_artifacts_json: bool = False,
    ) -> None:
        """
        Args:
            ignore_paths: The paths to ignore, see [`BaseFilter`][watchfiles.BaseFilter].
            extra_extensions: extra extensions to ignore.

        `ignore_paths` and `extra_extensions` can be passed as arguments partly to support [CLI](../cli.md) usage where
        `--ignore-paths` and `--extensions` can be passed as arguments.
        """
        self.extensions = (".py", ".pyx", ".pyd", ".pyc") + tuple(extra_extensions)
        self.include_artifacts_json = include_artifacts_json
        super().__init__(
            ignore_paths=ignore_paths,
            ignore_dirs=self.ignore_dirs + tuple([".databutton"]),
        )

    def __call__(self, change: "Change", path: str) -> bool:
        ret = (
            path.endswith(self.extensions)
            and super().__call__(change, path)
            and not path.endswith("artifacts.json")
        )
        if self.include_artifacts_json:
            ret = ret or path.endswith("artifacts.json")

        return ret


def get_components_hash():
    p = get_databutton_components_path()
    if not p.exists():
        return False
    md5 = hashlib.md5()
    with open(p, "r") as f:
        md5.update(f.read().encode("utf-8"))
    return md5.hexdigest()


class ComponentsJsonFilter(DefaultFilter):
    def __init__(self, starting_hash: str = None) -> None:
        super().__init__()
        self.prev_hash: Optional[str] = starting_hash

    def __call__(self, change: "Change", path: str) -> bool:
        should_call = super().__call__(change, path) and path.endswith("artifacts.json")
        if should_call:
            if not Path(path).exists():
                # Ignore if the file doesn't exist
                return False
            # Check hash extra check
            digest = get_components_hash()
            if digest == self.prev_hash:
                return False
            self.prev_hash = digest
            return True


class CouldNotStartError(Exception):
    pass


_streamlit_args = [
    ("server.address", "0.0.0.0"),
    ("server.enableWebsocketCompression", "true"),
    ("browser.gatherUsageStats", "false"),
    ("global.dataFrameSerialization", "arrow"),
    ("server.headless", "true"),
    ("server.enableCORS", "false"),
    ("server.enableXsrfProtection", "false"),
    ("server.runOnSave", "true"),
]


class StreamlitWatcher:
    def __init__(self):
        self.apps: Dict[str, StreamlitApp] = {}
        self.processes: Dict[str, Process] = {}

    def get_stable_app_id(self, app: StreamlitApp):
        return f"{app.uid}-{app.port}"

    async def clear(self):
        await self.cancel()
        self.apps = {}
        self.processes = {}

    async def cancel(self):
        async with create_task_group() as tg:
            for key, process in self.processes.items():
                tg.start_soon(self.stop_process, process, self.apps.get(key))

    async def start_process(self, uid: str, app: StreamlitApp) -> Process:
        logger.debug(f"Starting process for {app.name}")
        cmd = (
            f"streamlit run {app.filename} "
            + f"--server.port={app.port} "
            + " ".join([f"--{k}={v}" for k, v in _streamlit_args])
        )
        # Set environment and force PYTHONPATH
        current_env = os.environ.copy()
        current_env["PYTHONPATH"] = "."
        logger.debug(f"Starting:\n{cmd}")
        process = await open_process(cmd, env=current_env)
        logger.debug(f"Started streamlit process on {process.pid}")
        self.processes[uid] = process
        buffered = BufferedByteReceiveStream(process.stdout)
        try:
            streamlit_output = await buffered.receive_exactly(100)
            streamlit_output_str = str(repr(streamlit_output))
            if (
                "You can now view your Streamlit app in your browser"
                not in streamlit_output_str
            ):
                raise Exception("Error in starting streamlit")
            logger.debug(streamlit_output_str)
        except (anyio.EndOfStream, anyio.IncompleteRead):
            error = await process.stderr.receive()
            logger.info(f"Could not start streamlit app {app.name} on port {app.port}")
            logger.info(repr(error))
            raise CouldNotStartError(repr(error))

        return process

    async def stop_process(self, process: Process, app: StreamlitApp):
        try:
            logger.debug(f"Killing process {process.pid}")
            process.kill()
            await process.wait()
            logger.debug(f"Stopped process {process.pid}")
        except:  # noqa
            logger.debug(f"Could not terminate process for {process.pid}.")
            logger.debug(traceback.format_exc())
            # Ignore terminations, we'll nuke them all down below anyway

        # Streamlit has dangling processes, so let's find them and killem if need be
        for psprocess in psutil.process_iter():
            try:
                try:
                    cmdline = psprocess.cmdline()
                except (psutil.AccessDenied, psutil.NoSuchProcess, ProcessLookupError):
                    continue
                if app.filename in cmdline:
                    logger.debug(f"Killing psprocess {psprocess.pid}")
                    psprocess.kill()
                    psprocess.wait(2)
                    logger.debug(f"Forcefully stopped psutil.process {psprocess.pid}")
            except:  # noqa
                logger.debug(
                    f"Could not terminate psutil.process {psprocess.pid}. "
                    + f"{psprocess}",
                )

    async def update_processes_from_apps(
        self, apps: List[StreamlitApp]
    ) -> List[asyncio.Task]:
        apps_map: Dict[str, StreamlitApp] = {
            self.get_stable_app_id(app): app for app in apps
        }
        previous = self.apps
        previous_apps_map = self.apps.copy()
        self.apps = apps_map
        old, new = set(uid for uid in previous.keys()), set(
            uid for uid in apps_map.keys()
        )
        new_apps = list(new - old)
        deleted_apps = list(old - new)

        async with create_task_group() as tg:
            for new_uid in new_apps:
                tg.start_soon(self.start_process, new_uid, apps_map.get(new_uid))

            for deleted_uid in deleted_apps:
                tg.start_soon(
                    self.stop_process,
                    self.processes.get(deleted_uid),
                    previous_apps_map.get(deleted_uid),
                )

            for running_uid in new & old:
                new_app = apps_map.get(running_uid)
                old_app = previous_apps_map.get(running_uid)
                if old_app.uid != new_app.uid:
                    # This has a new port, we should restart it.
                    tg.start_soon(
                        self.stop_process, self.processes.get(running_uid), old_app
                    )
                    tg.start_soon(self.start_process, running_uid, new_app)

        return len(new_apps) > 0 or len(deleted_apps) > 0


@dataclass
class DatabuttonConfig:
    port: int = os.environ.get("PORT", 8000)
    log_level: str = os.environ.get("LOG_LEVEL", "critical")


class GracefulExit(SystemExit):
    code = 1


class DatabuttonRunner:
    def __init__(self, root_dir=Path.cwd(), **config):
        self.root_dir = root_dir
        self.config = DatabuttonConfig(**config)
        self.initial_hash: str = None

    async def create_webwatcher(self):
        args = [("port", self.config.port), ("log-level", self.config.log_level)]
        args_string = " ".join([f"--{arg}={value}" for arg, value in args])
        target_str = f"uvicorn {args_string} databutton.server.prod:app"
        return await arun_process(
            self.root_dir,
            target=target_str,
            target_type="command",
            watch_filter=ComponentsJsonFilter(starting_hash=self.initial_hash),
            callback=lambda _: click.secho("Restarting webserver..."),
        )

    async def create_streamlit_watcher(self):
        streamlit_watcher = StreamlitWatcher()
        self.streamlit_watcher = streamlit_watcher
        components = read_artifacts_json()
        await streamlit_watcher.update_processes_from_apps(components.streamlit_apps)

        async for _ in awatch(
            self.root_dir,
            watch_filter=ComponentsJsonFilter(starting_hash=self.initial_hash),
        ):
            logger.debug("Restarting streamlit apps...")
            new_components = read_artifacts_json()
            await streamlit_watcher.clear()
            await streamlit_watcher.update_processes_from_apps(
                new_components.streamlit_apps
            )

    async def create_scheduler_watcher(self):
        return await arun_process(
            self.root_dir,
            watch_filter=DatabuttonFilter(include_artifacts_json=True),
            target=Scheduler.create,
            callback=lambda _: click.secho("Restarting scheduler..."),
        )

    async def create_components_watcher(self):
        return await arun_process(
            self.root_dir,
            watch_filter=DatabuttonFilter(),
            target=partial(generate_components, self.root_dir),
        )

    async def signal_handler(self, scope: CancelScope):
        with open_signal_receiver(signal.SIGINT) as signals:
            async for _ in signals:
                click.secho("\nshutting down...", fg="cyan")
                await self.streamlit_watcher.clear()
                scope.cancel()
                logger.debug("asyncio scope cancelled")

    async def run(self, debug=False):
        if debug:
            logger.setLevel(logging.DEBUG)
        click.secho(
            f"starting databutton {click.style(f'v{__version__}', fg='green')}",
            fg="cyan",
        )
        shutil.rmtree(Path(".databutton"), ignore_errors=True)
        generate_components(self.root_dir)
        self.initial_hash = get_components_hash()
        try:
            async with create_task_group() as tg:
                tg.start_soon(self.signal_handler, tg.cancel_scope)
                tg.start_soon(self.create_streamlit_watcher, name="streamlit")
                tg.start_soon(self.create_components_watcher, name="components")
                tg.start_soon(self.create_scheduler_watcher, name="schedule")
                tg.start_soon(self.create_webwatcher, name="web")
        finally:
            logger.debug("Exited cleanly.")
